#include <ap_int.h>
#include <cstdlib>

#include "wino_macro.h"


#ifndef _WINO_IO_HPP_
#define _WINO_IO_HPP_

//template<int DUMMY>
void load_input_row_from_ddr(
		ap_uint<128>* DDR_port,
		ap_uint<16> input_buffer[16][INPUT_BUFFER_DEPTH],
		ap_uint<16> input_height,
		ap_uint<16> input_width,
		ap_uint<16> input_width_align_8,
		ap_uint<16> input_width_align_16,
		ap_uint<16> input_depth_aligned_8,
		ap_int<16> row_idx
)
{
	if(row_idx >= input_height) return;
	#pragma HLS interface m_axi port=DDR_port depth=200704
	#pragma HLS array_partition variable = input_buffer dim=1 complete

	ap_uint<16> register_matrix[8][8];
	#pragma HLS array_partition variable=register_matrix dim=1 complete
	#pragma HLS array_partition variable=register_matrix dim=2 complete

	int input_depth_div_8 = input_depth_aligned_8/8;
	int buffer_addr_offset;

	if(row_idx %16 <8)
		buffer_addr_offset = 0;
	else
		buffer_addr_offset = INPUT_BUFFER_DEPTH/2;

	ap_uint<1> increment_flag;
	int buffer_base_addr=0;

	for(int depth_tile_idx=0;depth_tile_idx<input_depth_div_8;depth_tile_idx++)
	{
		int ddr_address_offset = depth_tile_idx * input_height * input_width + row_idx*input_width;
		ap_uint<128>* DDR_offset = DDR_port + ddr_address_offset;
		ap_uint<3> depth_offset=0;
		ap_uint<1> left_down_flag=1;
		for(ap_uint<16> cycle=0;cycle<input_width_align_8;cycle++)
		{
		#pragma HLS pipeline
			ap_uint<128> DDR_data;

			DDR_data = DDR_offset[cycle];

			ap_uint<16> DDR_data_divided[8];
			#pragma HLS array_partition variable = DDR_data_divided complete

			if( cycle < input_width)
			{
				for(int i=0;i<8;i++)
				{
				#pragma HLS unroll
					DDR_data_divided[i]=DDR_data.range(i*16+15,i*16);
				}
			}
			else
			{
				for(int i=0;i<8;i++)
				{
				#pragma HLS unroll
					DDR_data_divided[i]=0;
				}
			}


			ap_uint<16> bank_data[8];
			#pragma HLS array_partition variable = bank_data complete

			if(left_down_flag)
			{
				for(int i=0;i<8;i++)
				{
				#pragma HLS unroll
					bank_data[i]=register_matrix[i][0];
				}
				for(int i=0;i<7;i++)
				{
				#pragma HLS unroll
					for(int j=0;j<8;j++)
					{
					#pragma HLS unroll
						register_matrix[j][i]=register_matrix[j][i+1];
					}
				}
				for(int i=0;i<8;i++)
				{ 
				#pragma HLS unroll
					register_matrix[i][7]=DDR_data_divided[i];
				}
			}
			else
			{
				for(int i=0;i<8;i++)
				{
				#pragma HLS unroll
					bank_data[i]=register_matrix[0][i];
				}

				for(int i=0;i<7;i++)
				{
				#pragma HLS unroll
					for(int j=0;j<8;j++)
					{
					#pragma HLS unroll
						register_matrix[i][j]=register_matrix[i+1][j];
					}
				}
				for(int i=0;i<8;i++)
				{ 
				#pragma HLS unroll
					register_matrix[7][i]=DDR_data_divided[i];
				}
			}

			if(cycle>=8)
			{
				input_buffer[left_down_flag*8+0][buffer_addr_offset+depth_offset]=bank_data[0];
				input_buffer[left_down_flag*8+1][buffer_addr_offset+depth_offset]=bank_data[1];
				input_buffer[left_down_flag*8+2][buffer_addr_offset+depth_offset]=bank_data[2];
				input_buffer[left_down_flag*8+3][buffer_addr_offset+depth_offset]=bank_data[3];
				input_buffer[left_down_flag*8+4][buffer_addr_offset+depth_offset]=bank_data[4];
				input_buffer[left_down_flag*8+5][buffer_addr_offset+depth_offset]=bank_data[5];
				input_buffer[left_down_flag*8+6][buffer_addr_offset+depth_offset]=bank_data[6];
				input_buffer[left_down_flag*8+7][buffer_addr_offset+depth_offset]=bank_data[7];
			}

			if(depth_offset==7)
				left_down_flag=~left_down_flag;

			if(cycle>=8 and cycle.range(3,0)==7)
				buffer_addr_offset+=8;
			depth_offset++;
		}

		for(ap_uint<16> cycle=0;cycle<8;cycle++)
		{
		#pragma HLS pipeline

			ap_uint<16> bank_data[8];
			#pragma HLS array_partition variable = bank_data complete
			if(left_down_flag)
			{
				for(int i=0;i<8;i++)
				{
				#pragma HLS unroll
					bank_data[i]=register_matrix[i][0];
				}
				for(int i=0;i<7;i++)
				{
				#pragma HLS unroll
					for(int j=0;j<8;j++)
					{
					#pragma HLS unroll
						register_matrix[j][i]=register_matrix[j][i+1];
					}
				}
			}
			else
			{
				for(int i=0;i<8;i++)
				{
				#pragma HLS unroll
					bank_data[i]=register_matrix[0][i];
				}

				for(int i=0;i<7;i++)
				{
				#pragma HLS unroll
					for(int j=0;j<8;j++)
					{
					#pragma HLS unroll
						register_matrix[i][j]=register_matrix[i+1][j];
					}
				}
			}

			input_buffer[left_down_flag*8+0][buffer_addr_offset+depth_offset]=bank_data[0];
			input_buffer[left_down_flag*8+1][buffer_addr_offset+depth_offset]=bank_data[1];
			input_buffer[left_down_flag*8+2][buffer_addr_offset+depth_offset]=bank_data[2];
			input_buffer[left_down_flag*8+3][buffer_addr_offset+depth_offset]=bank_data[3];
			input_buffer[left_down_flag*8+4][buffer_addr_offset+depth_offset]=bank_data[4];
			input_buffer[left_down_flag*8+5][buffer_addr_offset+depth_offset]=bank_data[5];
			input_buffer[left_down_flag*8+6][buffer_addr_offset+depth_offset]=bank_data[6];
			input_buffer[left_down_flag*8+7][buffer_addr_offset+depth_offset]=bank_data[7];



			if(depth_offset==7)
				left_down_flag=~left_down_flag;
			// if(cycle.range(3,0)==7)
			// 	buffer_addr_offset+=8;
			depth_offset++;
		}
		buffer_addr_offset+=8;
	}
}




//template<int DUMMY>
void load_input_rowtile_from_ddr(
		ap_uint<128>* DDR_port1,
        ap_uint<128>* DDR_port2,
        ap_uint<128>* DDR_port3,
        ap_uint<128>* DDR_port4,
		ap_uint<16> input_buffer[8][16][INPUT_BUFFER_DEPTH],
		ap_uint<16> input_height,
		ap_uint<16> input_width,
		ap_uint<16> input_width_align_8,
		ap_uint<16> input_width_align_16,
		ap_uint<16> input_depth_aligned_8,
        ap_uint<16> output_height,
        ap_uint<16> next_start_output_row,
        ap_uint<16> pad_size,
        ap_uint<1> first_flag
)
{
	#pragma HLS array_partition variable =input_buffer dim=1 complete
	#pragma HLS array_partition variable =input_buffer dim=2 complete
    static int loaded_input_row_number;
    if(first_flag)
        loaded_input_row_number = 0;
    
    int required_loaded_input_row_number = next_start_output_row - pad_size + WINO_SIZE;

    if ( required_loaded_input_row_number > output_height )
        required_loaded_input_row_number = output_height;
    while( required_loaded_input_row_number > loaded_input_row_number)
    {
        if( loaded_input_row_number%8/4 == 0)
        {

            load_input_row_from_ddr(
                DDR_port1,
                input_buffer[0],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number
            );


            load_input_row_from_ddr(
                DDR_port2,
                input_buffer[1],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number+1
            );


            load_input_row_from_ddr(
                DDR_port3,
                input_buffer[2],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number+2
            );

            load_input_row_from_ddr(
                DDR_port4,
                input_buffer[3],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number+3
            );

        }
        else
        {

            load_input_row_from_ddr(
                DDR_port1,
                input_buffer[4],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number
            );

            load_input_row_from_ddr(
                DDR_port2,
                input_buffer[5],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number+1
            );


            load_input_row_from_ddr(
                DDR_port3,
                input_buffer[6],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number+2
            );

            load_input_row_from_ddr(
                DDR_port4,
                input_buffer[7],
                input_height,
                input_width,
				input_width_align_8,
				input_width_align_16,
                input_depth_aligned_8,
                loaded_input_row_number+3
            );
        }
        loaded_input_row_number+=4;
    }
}



//template<int DUMMY>
void write_output_to_DDR(
	ap_uint<128>* DDR_port0,
	ap_uint<128>* DDR_port1,
	ap_uint<128>* DDR_port2,
	ap_uint<128>* DDR_port3,
	ap_uint<36> output_buffer[16][16][1024],
	int outdepth_ceildiv8,
	int start_out_row_idx,
	int end_row,
	int out_height_aligned4,	
	int out_width,
	int wino_output_tile_size,
	int row_tile_number,
	int wino_flag,
	int final_right_shift)
{
#pragma HLS interface m_axi port = DDR_port0
#pragma HLS interface m_axi port = DDR_port1
#pragma HLS interface m_axi port = DDR_port2
#pragma HLS interface m_axi port = DDR_port3

#pragma HLS array_partition variable = output_buffer dim=1 complete
#pragma HLS array_partition variable = output_buffer dim=2 complete

	ap_uint<1> valid_output_flag[4];
	
	int row_offset0 = (start_out_row_idx+0)*out_width; 
	int row_offset1 = (start_out_row_idx+1)*out_width; 
	int row_offset2 = (start_out_row_idx+2)*out_width; 
	int row_offset3 = (start_out_row_idx+3)*out_width;

	int plane_offset_step = out_width*out_height_aligned4;
	int depth_base_offset = 0;
	

	for(int depth_idx=0;depth_idx<outdepth_ceildiv8;depth_idx++)
	{
		
		ap_uint<128>* DDR_offset0= DDR_port0+depth_idx*out_width*out_height_aligned4+row_offset0;
		ap_uint<128>* DDR_offset1= DDR_port1+depth_idx*out_width*out_height_aligned4+row_offset1;
		ap_uint<128>* DDR_offset2= DDR_port2+depth_idx*out_width*out_height_aligned4+row_offset2;
		ap_uint<128>* DDR_offset3= DDR_port3+depth_idx*out_width*out_height_aligned4+row_offset3;

			
		for(ap_int<16> out_col_idx=0;out_col_idx<out_width;out_col_idx++)
		{
#pragma HLS pipeline
			int address = depth_idx*row_tile_number + out_col_idx/(wino_output_tile_size*2);
			ap_uint<36> output_buffer_value_wire[8][4][2][4];
#pragma HLS array_partition variable = output_buffer_value_wire  complete
			for(int i=0;i<8;i++){
			#pragma HLS unroll	
			for(int j=0;j<2;j++){
			#pragma HLS unroll	
			for(int k=0;k<4;k++){
			#pragma HLS unroll	
			for(int l=0;l<4;l++){
			#pragma HLS unroll	
				output_buffer_value_wire[i][k][j][l]=output_buffer[j*8+2][k*4+l][address];
			}
			}
			}
			}

			

			ap_uint<36> write_back_wire[8][4];
#pragma HLS array_partition variable = write_back_wire  complete
			ap_uint<2> pix_sel = out_col_idx.range(1,0);
			ap_uint<1> tile_sel = out_col_idx[2];

			for(int i=0;i<8;i++){
			#pragma HLS unroll	
			for(int j=0;j<4;j++){
			#pragma HLS unroll	
				write_back_wire[i][j]=output_buffer_value_wire[i][j][tile_sel][pix_sel];
			}
			}

			ap_int<8> write_back_wire_batch0[8][4];
			ap_int<8> write_back_wire_batch1[8][4];
#pragma HLS array_partition variable = write_back_wire_batch0  complete
#pragma HLS array_partition variable = write_back_wire_batch1  complete
			for(int i=0;i<8;i++){
				#pragma HLS unroll
			for(int j=0;j<4;j++){
				#pragma HLS unroll
				write_back_wire_batch0[i][j]= (write_back_wire[i][j].range(17,0)) >> final_right_shift;
				write_back_wire_batch1[i][j]= (write_back_wire[i][j].range(35,18)) >> final_right_shift;
			}
			}

			ap_uint<128> write_back_128bit3;
			ap_uint<128> write_back_128bit2;
			ap_uint<128> write_back_128bit1;
			ap_uint<128> write_back_128bit0;

			for(int i=0;i<8;i++)
			{
				#pragma HLS unroll
				write_back_128bit0.range(i*16+15,i*16) = ( write_back_wire_batch1[i][0],write_back_wire_batch0[i][0]);
				write_back_128bit1.range(i*16+15,i*16) = ( write_back_wire_batch1[i][1],write_back_wire_batch0[i][1]);
				write_back_128bit2.range(i*16+15,i*16) = ( write_back_wire_batch1[i][2],write_back_wire_batch0[i][2]);
				write_back_128bit3.range(i*16+15,i*16) = ( write_back_wire_batch1[i][3],write_back_wire_batch0[i][3]);
			}

			
			DDR_offset3[out_col_idx]=write_back_128bit3;
			DDR_offset2[out_col_idx]=write_back_128bit2;
			DDR_offset1[out_col_idx]=write_back_128bit1;
			DDR_offset0[out_col_idx]=write_back_128bit0;

		}
	}

}



#endif
